import inspect
import logging
import os
import textwrap
import typing as t
from collections import deque, namedtuple
from pprint import pformat

from django_docker_helpers.utils import import_from, shred, wf, run_env_once
from . import exceptions
from .backends import *

DEFAULT_PARSER_MODULE_PATH = 'django_docker_helpers.config.backends'

DEFAULT_PARSER_MODULES = (
    '{0}.EnvironmentParser'.format(DEFAULT_PARSER_MODULE_PATH),
    '{0}.MPTRedisParser'.format(DEFAULT_PARSER_MODULE_PATH),
    '{0}.MPTConsulParser'.format(DEFAULT_PARSER_MODULE_PATH),
    '{0}.RedisParser'.format(DEFAULT_PARSER_MODULE_PATH),
    '{0}.ConsulParser'.format(DEFAULT_PARSER_MODULE_PATH),
    '{0}.YamlParser'.format(DEFAULT_PARSER_MODULE_PATH),
)


def comma_str_to_list(raw_val: str) -> t.List[str]:
    return list(filter(None, raw_val.split(',')))


ConfigReadItem = namedtuple('ConfigReadItem', ['variable_path', 'value', 'type', 'is_default', 'parser_name'])


class ConfigLoader:
    """
    - provides a single interface to read from specified config parsers in order they present;
    - tracks accessed from parsers options;
    - prints config options access log in pretty-print way.

    Example:
        ::

            env = {
                'PROJECT__DEBUG': 'false'
            }
            parsers = [
                EnvironmentParser(scope='project', env=env),
                RedisParser('my/conf/service/config.yml', host=REDIS_HOST, port=REDIS_PORT),
                YamlParser(config='./tests/data/config.yml', scope='project'),
            ]
            configure = ConfigLoader(parsers=parsers)

            DEBUG = configure('debug')  # 'false'
            DEBUG = configure('debug', coerce_type=bool)  # False
    """
    def __init__(self,
                 parsers: t.List[BaseParser],
                 silent: bool = False,
                 suppress_logs: bool = False,
                 keep_read_records_max: int = 1024):
        """
        Initialization:
            - takes a list of initialized parsers;
            - it's supposed to use ONLY unique parsers for ``parsers`` argument
              (or you are going to get the same initial arguments for all parsers of the same type in
              :meth:`~django_docker_helpers.config.ConfigLoader.from_env`);
            - the parsers's order does matter.

        :param parsers: a list of initialized parsers
        :param silent: don't raise exceptions if any read attempt failed
        :param suppress_logs: don't display any exception warnings on screen
        :param keep_read_records_max: max capacity queue length
        """
        self.parsers = parsers
        self.silent = silent
        self.suppress_logs = suppress_logs
        self.sentinel = object()
        self.logger = logging.getLogger(self.__class__.__name__)
        self.config_read_queue = deque(maxlen=keep_read_records_max)

        self.colors_map = {
            'title': '\033[1;35m',

            'parser': '\033[0;33m',
            'path': '\033[94m',
            'type': '\033[1;33m',
            'value': '\033[32m',

            'reset': '\033[0m',
        }

    def enqueue(self,
                variable_path: str,
                parser: t.Optional[BaseParser] = None,
                value: t.Any = None):
        self.config_read_queue.append(ConfigReadItem(
            variable_path,
            shred(variable_path, value),
            type(value).__name__,
            not bool(parser),
            str(parser),
        ))

    def __call__(self,
                 variable_path: str,
                 default: t.Optional[t.Any] = None,
                 coerce_type: t.Optional[t.Type] = None,
                 coercer: t.Optional[t.Callable] = None,
                 required: bool = False,
                 **kwargs):
        """
        A useful shortcut for method :meth:`~django_docker_helpers.config.ConfigLoader.get`
        """
        return self.get(
            variable_path, default=default,
            coerce_type=coerce_type, coercer=coercer,
            required=required,
            **kwargs)

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            required: bool = False,
            **kwargs):
        """
        Tries to read a ``variable_path`` from each of the passed parsers.
        It stops if read was successful and returns a retrieved value.
        If none of the parsers contain a value for the specified path it returns ``default``.

        :param variable_path: a path to variable in config
        :param default: a default value if ``variable_path`` is not present anywhere
        :param coerce_type: cast a result to a specified type
        :param coercer: perform the type casting with specified callback
        :param required: raise ``RequiredValueIsEmpty`` if no ``default`` and no result
        :param kwargs: additional options to all parsers
        :return: **the first successfully read** value from the list of parser instances or ``default``

        :raises config.exceptions.RequiredValueIsEmpty: if nothing is read,``required``
         flag is set, and there's no ``default`` specified
        """

        for p in self.parsers:
            try:
                val = p.get(
                    variable_path, default=self.sentinel,
                    coerce_type=coerce_type, coercer=coercer,
                    **kwargs
                )
                if val != self.sentinel:
                    self.enqueue(variable_path, p, val)
                    return val
            except Exception as e:
                if not self.silent:
                    raise
                if self.suppress_logs:
                    continue
                self.logger.error('Parser {0} cannot get key `{1}`: {2}'.format(
                    p.__class__.__name__,
                    variable_path,
                    str(e)
                ))

        self.enqueue(variable_path, value=default)

        if not default and required:
            raise exceptions.RequiredValueIsEmpty(
                'No default provided and no value read for `{0}`'.format(variable_path))

        return default

    @staticmethod
    def import_parsers(parser_modules: t.Iterable[str]) -> t.Generator[t.Type[BaseParser], None, None]:
        """
        Resolves and imports all modules specified in ``parser_modules``. Short names from the local scope
        are supported (the scope is ``django_docker_helpers.config.backends``).

        :param parser_modules: a list of dot-separated module paths
        :return: a generator of [probably] :class:`~django_docker_helpers.config.backends.base.BaseParser`

        Example:
        ::

            parsers = list(ConfigLoader.import_parsers([
                'EnvironmentParser',
                'django_docker_helpers.config.backends.YamlParser'
            ]))
            assert parsers == [EnvironmentParser, YamlParser]
        """
        for import_path in parser_modules:
            path_parts = import_path.rsplit('.', 1)
            if len(path_parts) == 2:
                mod_path, parser_class_name = path_parts
            else:
                mod_path = DEFAULT_PARSER_MODULE_PATH
                parser_class_name = import_path

            yield import_from(mod_path, parser_class_name)

    @staticmethod
    def load_parser_options_from_env(
            parser_class: t.Type[BaseParser],
            env: t.Optional[t.Dict[str, str]] = None) -> t.Dict[str, t.Any]:
        """
        Extracts arguments from ``parser_class.__init__`` and populates them from environment variables.

        Uses ``__init__`` argument type annotations for correct type casting.

        .. note::

            Environment variables should be prefixed with ``<UPPERCASEPARSERCLASSNAME>__``.

        :param parser_class: a subclass of :class:`~django_docker_helpers.config.backends.base.BaseParser`
        :param env: a dict with environment variables, default is ``os.environ``
        :return: parser's ``__init__`` arguments dict mapping

        Example:
        ::

            env = {
                'REDISPARSER__ENDPOINT': 'go.deep',
                'REDISPARSER__HOST': 'my-host',
                'REDISPARSER__PORT': '66',
            }

            res = ConfigLoader.load_parser_options_from_env(RedisParser, env)
            assert res == {'endpoint': 'go.deep', 'host': 'my-host', 'port': 66}
        """
        env = env or os.environ
        sentinel = object()
        spec: inspect.FullArgSpec = inspect.getfullargspec(parser_class.__init__)
        environment_parser = EnvironmentParser(scope=parser_class.__name__.upper(), env=env)

        stop_args = ['self']
        safe_types = [int, bool, str]

        init_args = {}

        for arg_name in spec.args:
            if arg_name in stop_args:
                continue

            type_hint = spec.annotations.get(arg_name)
            coerce_type = None

            if type_hint in safe_types:
                coerce_type = type_hint
            elif hasattr(type_hint, '__args__'):
                if len(type_hint.__args__) == 1:  # one type
                    if type_hint.__args__[0] in safe_types:
                        coerce_type = type_hint.__args__[0]
                elif len(type_hint.__args__) == 2:  # t.Optional
                    try:
                        _args = list(type_hint.__args__)
                        _args.remove(type(None))
                        if _args[0] in safe_types:
                            coerce_type = _args[0]
                    except ValueError:
                        pass

            val = environment_parser.get(arg_name, sentinel, coerce_type=coerce_type)
            if val is sentinel:
                continue

            init_args[arg_name] = val

        return init_args

    @staticmethod
    def from_env(parser_modules: t.Optional[t.Union[t.List[str], t.Tuple[str]]] = DEFAULT_PARSER_MODULES,
                 env: t.Optional[t.Dict[str, str]] = None,
                 silent: bool = False,
                 suppress_logs: bool = False,
                 extra: t.Optional[dict] = None) -> 'ConfigLoader':
        """
        Creates an instance of :class:`~django_docker_helpers.config.ConfigLoader`
        with parsers initialized from environment variables.

        By default it tries to initialize all bundled parsers.
        Parsers may be customized with ``parser_modules`` argument or ``CONFIG__PARSERS`` environment variable.
        Environment variable has a priority over the method argument.

        :param parser_modules: a list of dot-separated module paths
        :param env: a dict with environment variables, default is ``os.environ``
        :param silent: passed to :class:`~django_docker_helpers.config.ConfigLoader`
        :param suppress_logs: passed to :class:`~django_docker_helpers.config.ConfigLoader`
        :param extra: pass extra arguments to *every* parser
        :return: an instance of :class:`~django_docker_helpers.config.ConfigLoader`

        Example:
        ::

            env = {
                'CONFIG__PARSERS': 'EnvironmentParser,RedisParser,YamlParser',
                'ENVIRONMENTPARSER__SCOPE': 'nested',
                'YAMLPARSER__CONFIG': './tests/data/config.yml',
                'REDISPARSER__HOST': 'wtf.test',
                'NESTED__VARIABLE': 'i_am_here',
            }

            loader = ConfigLoader.from_env(env=env)
            assert [type(p) for p in loader.parsers] == [EnvironmentParser, RedisParser, YamlParser]
            assert loader.get('variable') == 'i_am_here', 'Ensure env copied from ConfigLoader'

            loader = ConfigLoader.from_env(parser_modules=['EnvironmentParser'], env={})

        """
        env = env or os.environ
        extra = extra or {}
        environment_parser = EnvironmentParser(scope='config', env=env)
        silent = environment_parser.get('silent', silent, coerce_type=bool)
        suppress_logs = environment_parser.get('suppress_logs', suppress_logs, coerce_type=bool)

        env_parsers = environment_parser.get('parsers', None, coercer=comma_str_to_list)
        if not env_parsers and not parser_modules:
            raise ValueError('Must specify `CONFIG__PARSERS` env var or `parser_modules`')

        if env_parsers:
            parser_classes = ConfigLoader.import_parsers(env_parsers)
        else:
            parser_classes = ConfigLoader.import_parsers(parser_modules)

        parsers = []

        for parser_class in parser_classes:
            parser_options = ConfigLoader.load_parser_options_from_env(parser_class, env=env)

            _init_args = inspect.getfullargspec(parser_class.__init__).args
            # add extra args if parser's __init__ can take it it
            if 'env' in _init_args:
                parser_options['env'] = env

            for k, v in extra.items():
                if k in _init_args:
                    parser_options[k] = v

            parser_instance = parser_class(**parser_options)
            parsers.append(parser_instance)

        return ConfigLoader(parsers=parsers, silent=silent, suppress_logs=suppress_logs)

    def _colorize(self, name: str, value: str, use_color: bool = False) -> str:
        if not use_color:
            return value
        color = self.colors_map.get(name, '')
        if not color:
            return value
        reset = self.colors_map['reset']
        parts = [color + p + reset for p in str(value).split('\n')]
        return '\n'.join(parts)

    @staticmethod
    def _pformat(raw_obj: t.Union[str, t.Any], width: int = 50) -> str:
        raw_str = str(raw_obj)
        if len(raw_str) <= width:
            return raw_obj
        if isinstance(raw_obj, str):
            return '\n'.join(textwrap.wrap(raw_str, width=width))
        return pformat(raw_obj, width=width, compact=True)

    @run_env_once
    def print_config_read_queue(
            self,
            use_color: bool = False,
            max_col_width: int = 50):
        """
        Prints all read (in call order) options.

        :param max_col_width: limit column width, ``50`` by default
        :param use_color: use terminal colors
        :return: nothing
        """
        wf(self.format_config_read_queue(use_color=use_color, max_col_width=max_col_width))
        wf('\n')

    def format_config_read_queue(self,
                                 use_color: bool = False,
                                 max_col_width: int = 50) -> str:
        """
        Prepares a string with pretty printed config read queue.

        :param use_color: use terminal colors
        :param max_col_width: limit column width, ``50`` by default
        :return:
        """
        try:
            from terminaltables import SingleTable
        except ImportError:
            import warnings
            warnings.warn('Cannot display config read queue. Install terminaltables first.')
            return ''

        col_names_order = ['path', 'value', 'type', 'parser']
        pretty_bundles = [[self._colorize(name, name.capitalize(), use_color=use_color)
                           for name in col_names_order]]

        for config_read_item in self.config_read_queue:
            pretty_attrs = [
                config_read_item.variable_path,
                config_read_item.value,
                config_read_item.type,
                config_read_item.parser_name
            ]
            pretty_attrs = [self._pformat(pa, max_col_width) for pa in pretty_attrs]

            if config_read_item.is_default:
                pretty_attrs[0] = '*' + pretty_attrs[0]

            if use_color:
                pretty_attrs = [self._colorize(column_name, pretty_attr, use_color=use_color)
                                for column_name, pretty_attr in zip(col_names_order, pretty_attrs)]
            pretty_bundles.append(pretty_attrs)

        table = SingleTable(pretty_bundles)
        table.title = self._colorize('title', 'CONFIG READ QUEUE', use_color=use_color)
        table.justify_columns[0] = 'right'
        # table.inner_row_border = True
        return str(table.table)



from .base import BaseParser
from .consul_parser import ConsulParser
from .environment_parser import EnvironmentParser
from .mpt_consul_parser import MPTConsulParser
from .mpt_redis_parser import MPTRedisParser
from .redis_parser import RedisParser
from .yaml_parser import YamlParser

__all__ = [
    'BaseParser',

    'YamlParser',
    'EnvironmentParser',

    'MPTConsulParser',
    'MPTRedisParser',

    'ConsulParser',
    'RedisParser',
]



from json import loads as json_load

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
from yaml import load as yaml_load

from django_docker_helpers.config.backends.environment_parser import EnvironmentParser

pytestmark = [pytest.mark.backend, pytest.mark.env]


# noinspection PyMethodMayBeStatic
class EnvironmentBackendTest:
    def test__get_env_var_name(self):
        parser = EnvironmentParser('project')
        assert parser.get_env_var_name('my.nested.variable') == 'PROJECT__MY__NESTED__VARIABLE'
        parser = EnvironmentParser()
        assert parser.get_env_var_name('my.nested.variable') == 'MY__NESTED__VARIABLE'
        parser = EnvironmentParser(nested_delimiter='_NOPLEASE_')
        assert parser.get_env_var_name('my.nested.variable') == 'MY_NOPLEASE_NESTED_NOPLEASE_VARIABLE'

        parser = EnvironmentParser('my.long.project')
        assert parser.get_env_var_name('variable') == 'MY__LONG__PROJECT__VARIABLE'

    def test__get(self):
        env = {
            'MY__VARIABLE': '33',
            'MY__NESTED__YAML__LIST__VARIABLE': '[33, 42]',
            'MY__NESTED__JSON__LIST__VARIABLE': '["33", 42]',
            'MY__NESTED__JSON__DICT__VARIABLE': '{"obj": true}',
            'MY_STR_BOOL_TRUE': 'true',
            'MY_STR_BOOL_FALSE': 'false',
            'MY_INT_BOOL_TRUE': '1',
            'MY_INT_BOOL_False': '0',
        }
        parser = EnvironmentParser(env=env)
        assert parser.get('my.variable') == '33', 'Ensure return the same value if no coercing'
        assert parser.get('my.variable', coerce_type=str) == '33', 'Ensure str coercing'
        assert parser.get('my.variable', coerce_type=int) == 33, 'Ensure int coercing'

        # test complex types
        parser = EnvironmentParser(env=env)
        assert parser.get('my.nested.yaml.list.variable', coerce_type=list, coercer=yaml_load) == [33, 42]
        assert parser.get('my.nested.json.list.variable', coerce_type=list, coercer=json_load) == ['33', 42]

        assert parser.get('my.nested.json.dict.variable', coerce_type=dict, coercer=json_load) == {'obj': True}
        assert parser.get('my.nested.json.dict.variable', coercer=json_load) == {'obj': True}
        assert parser.get('my.nested.json.dict.variable', coercer=json_load, coerce_type=dict) == {'obj': True}

        # test dot-separated scope
        parser = EnvironmentParser(env=env, scope='my.nested')
        assert parser.get('yaml.list.variable', coerce_type=list, coercer=yaml_load) == [33, 42]

    def test__path_separator(self):
        _env = {
            'MY__VARIABLE': '33',
        }
        p = EnvironmentParser(env=_env, path_separator='/')
        assert p.get('my/variable', coerce_type=int) == 33

    def test__get__bool_coercing(self):
        _env = {
            'MY_STR_BOOL_TRUE': 'true',
            'MY_STR_BOOL_FALSE': 'false',
            'MY_INT_BOOL_TRUE': '1',
            'MY_INT_BOOL_FALSE': '0',
        }
        p = EnvironmentParser(env=_env)

        assert p.get('my_str_bool_true', coerce_type=bool) is True
        assert p.get('my_str_bool_false', coerce_type=bool) is False

        assert p.get('my_int_bool_true', coerce_type=bool) is True
        assert p.get('my_int_bool_false', coerce_type=bool) is False

    def test__get_client__raises(self):
        p = EnvironmentParser(env={}, path_separator='/')
        with pytest.raises(NotImplementedError):
            p.get_client()



import os
import typing as t

from django.core.management.base import BaseCommand

from django_docker_helpers.config import ConfigLoader


def write_uwsgi_ini_cfg(fp: t.IO, cfg: dict):
    """
    Writes into IO stream the uwsgi.ini file content (actually it does smth strange, just look below).

    uWSGI configs are likely to break INI (YAML, etc) specification (double key definition)
    so it writes `cfg` object (dict) in "uWSGI Style".

    >>> import sys
    >>> cfg = {
    ... 'static-map': [
    ... '/static/=/application/static/',
    ... '/media/=/application/media/',
    ... '/usermedia/=/application/usermedia/']
    ... }
    >>> write_uwsgi_ini_cfg(sys.stdout, cfg)
    [uwsgi]
    static-map = /static/=/application/static/
    static-map = /media/=/application/media/
    static-map = /usermedia/=/application/usermedia/
    """
    fp.write(f'[uwsgi]\n')

    for key, val in cfg.items():
        if isinstance(val, bool):
            val = str(val).lower()

        if isinstance(val, list):
            for v in val:
                fp.write(f'{key} = {v}\n')
            continue

        fp.write(f'{key} = {val}\n')


class Command(BaseCommand):
    help = 'Runs uWSGI with an INI config fetched from ConfigLoader'

    def add_arguments(self, parser):
        parser.add_argument(
            '--key', default='uwsgi', dest='key',
            help='The YAML configuration key with which uWSGI config can be accessed (default: uwsgi)'
        )
        parser.add_argument(
            '--print', action='store_true', dest='print',
            help='If true the ini file content will be outputted'
        )
        parser.add_argument(
            '-f', '--file', default='/tmp/uwsgi.ini', dest='file',
            help='Specifies a file that will be used for config storing'
        )

    def handle(self, **options):
        key = options['key']

        configure = ConfigLoader.from_env(suppress_logs=True, silent=True)
        uwsgi_cfg = configure.get(key)

        if not uwsgi_cfg:
            self.stderr.write(f'Parsing error: uWSGI config was not found by the key "{key}"')
            return

        if options['print']:
            self.stdout.write('*' * 80 + '\n')
            self.stdout.write('uWSGI CONFIG'.center(80))
            self.stdout.write('*' * 80 + '\n')

            write_uwsgi_ini_cfg(self.stdout, uwsgi_cfg)

            self.stdout.write('*' * 80 + '\n')
            self.stdout.flush()

        cfg_file = options['file']
        with open(cfg_file, 'w') as file:
            write_uwsgi_ini_cfg(file, uwsgi_cfg)

        os.execvp('uwsgi', ('--ini', cfg_file))



import os

# noinspection PyPackageRequirements
import pytest

from django_docker_helpers import utils

pytestmark = pytest.mark.utils


# noinspection PyMethodMayBeStatic
class UtilsTest:
    def test__utils__dotkey(self):
        obj = {
            'debug': True,
            'first': 1,
            'second': {
                'nested': True,
                'next': [1, 2, 3]
            }
        }
        assert utils.dotkey(obj, 'debug') is True
        assert utils.dotkey(obj, 'first') == 1
        assert utils.dotkey(obj, 'second') == {'nested': True, 'next': [1, 2, 3]}
        assert utils.dotkey(obj, 'second.next.1') == 2
        assert utils.dotkey(obj, 'second.lol', 'DEFAULT') == 'DEFAULT'

    def test__coerce_str_to_bool(self):
        assert utils.coerce_str_to_bool('0') is False
        assert utils.coerce_str_to_bool('1') is True

        assert utils.coerce_str_to_bool('FALSE') is False
        assert utils.coerce_str_to_bool('true') is True

        assert utils.coerce_str_to_bool('off') is False
        assert utils.coerce_str_to_bool('on') is True

        assert utils.coerce_str_to_bool(False) is False
        assert utils.coerce_str_to_bool(True) is True

        assert utils.coerce_str_to_bool(0) is False
        assert utils.coerce_str_to_bool(1) is True

        assert utils.coerce_str_to_bool('qwe') is True

        with pytest.raises(ValueError):
            utils.coerce_str_to_bool('lol', strict=True)

    def test__utils__env_bool_flag(self):
        _env = {
            'test_int_0': 0,
            'test_int_1': 1,
            'test_int_2': 2,
            'test_str_0': '0',
            'test_str_1': '1',
            'test_str_2': '2',
            'test_str_true': 'true',
            'test_str_false': 'false',
            'test_bool_true': True,
            'test_bool_false': False,
        }
        assert utils.env_bool_flag('test_int_0', env=_env) is False
        assert utils.env_bool_flag('test_int_1', env=_env) is True
        assert utils.env_bool_flag('test_int_2', env=_env) is True
        with pytest.raises(ValueError):
            assert utils.env_bool_flag('test_int_2', strict=True, env=_env)

        assert utils.env_bool_flag('test_str_0', env=_env) is False
        assert utils.env_bool_flag('test_str_1', env=_env) is True
        assert utils.env_bool_flag('test_str_2', env=_env) is True
        with pytest.raises(ValueError):
            assert utils.env_bool_flag('test_str_2', strict=True, env=_env)

        assert utils.env_bool_flag('test_str_true', env=_env) is True
        assert utils.env_bool_flag('test_str_false', env=_env) is False

        assert utils.env_bool_flag('test_bool_true', env=_env) is True
        assert utils.env_bool_flag('test_bool_false', env=_env) is False

        assert utils.env_bool_flag('doesnotexist', env=_env) is False

    def test__utils__materialize_dict(self):
        # check yield
        sample = {
            'plain': 1,
            'a_some': {
                'path': 1
            },
            'and': {
                1: 2,
                'more': [1, 2, 3],
                'much': {
                    'a': 1,
                    'more': 2,
                    'really': {
                        'longer': {
                            'path': 666
                        }
                    }
                },
                'bad': [1, {'no': 'please'}]
            }
        }
        results = [
            ('and/much/really/longer/path', 666),
            ('and/much/a', 1),
            ('and/much/more', 2),
            ('a_some/path', 1),
            ('and/1', 2),
            ('and/bad', [1, {'no': 'please'}]),
            ('and/more', [1, 2, 3]),
            ('plain', 1)
        ]

        for gen_val in utils._materialize_dict(sample, separator='/'):
            assert gen_val in results, 'Ensure transformed path is present in results'

        for gen_val, reference_val in zip(utils.materialize_dict(sample, separator='/'), results):
            assert gen_val == reference_val, 'Ensure ordering is correct'

    def test__utils__mp_serialize_dict(self):
        sample = {
            'bool_flag': '',  # flag
            'unicode': 'вася',
            'none_value': None,
            'debug': True,
            'mixed': ['ascii', 'юникод', 1, {'d': 1}, {'b': 2}],
            'nested': {
                'a': {
                    'b': 2,
                    'c': b'bytes',
                }
            }
        }

        result = [
            ('nested/a/b', b'2'),
            ('nested/a/c', b'bytes'),
            ('bool_flag', b"::YAML::\n''\n"),
            ('debug', b'true'),
            ('mixed', b'::YAML::\n- ascii\n- '
                      b'"\\u044E\\u043D\\u0438\\u043A\\u043E\\u0434"\n- 1\n- '
                      b'{d: 1}\n- {b: 2}\n'),
            ('none_value', None),
            ('unicode', b'\xd0\xb2\xd0\xb0\xd1\x81\xd1\x8f')
        ]

        md = utils.mp_serialize_dict(sample, separator='/')
        assert md == result

    def test__run_env_once(self):
        res = []

        @utils.run_env_once
        def only_once():
            res.append(1)

        only_once()
        only_once()

        assert len(res) == 1
        assert os.getenv('only_once') == '1'

    def test__dot_path(self):
        class O1:
            my_dict = {'a': {'b': 1}}

        class O2:
            def __init__(self):
                self.nested = O1()

        class O3:
            final = O2()

        o = O3()

        assert utils.dot_path(o, 'final.nested.my_dict.a.b') == 1
        assert utils.dot_path(o, 'final.lol.qwe', 'my_default') == 'my_default'
        assert utils.dot_path(o, 'final.nested.my_dict.a.qwe', 'my_default') == 'my_default'

    def test__shred(self):
        assert utils.shred('password', '1234') == '****'
        assert utils.shred('qwerty', '1234') == '1234'

    def test__shred_deep(self):
        shredded = utils.shred_deep({
            'password': 1,
            'password_2': {
                'password': 4444,
            },
            'list': [
                {'password': 222},
                'normal',
                {'normal': 1}
            ],
            'set': {1, 2, 3},
            'frozenset': frozenset([1, 2, 3, 'password'])
        })

        assert shredded == {
            'password': '*',
            'password_2': {'password': '****'},
            'list': [
                {'password': '***'},
                'normal',
                {'normal': 1}],
            'set': {1, 2, 3},
            'frozenset': frozenset({1, 2, 3, 'password'})
        }



# noinspection PyPackageRequirements
import pytest

from django_docker_helpers.config.backends import BaseParser

pytestmark = [pytest.mark.backend, pytest.mark.base]


# noinspection PyMethodMayBeStatic
class BaseParserTest:
    def test__base_parser__raises__not_implemented(self):
        p = BaseParser()
        with pytest.raises(NotImplementedError):
            p.get('qwe')

        with pytest.raises(NotImplementedError):
            p.get_client()



# noinspection PyPackageRequirements
import pytest

from django_docker_helpers.config.backends.yaml_parser import YamlParser

pytestmark = [pytest.mark.backend, pytest.mark.yaml]


# noinspection PyMethodMayBeStatic
class YamlBackendTest:
    def test__yaml_parser__init(self):
        # no config
        with pytest.raises(ValueError):
            YamlParser()

        p = YamlParser('./tests/data/config.yml')
        assert isinstance(p.data, dict)

    def test__yaml_parser__get(self):
        p = YamlParser('./tests/data/config.yml')
        assert p.get('debug') is True
        assert p.get('my.deep.nested.variable') == 1
        assert p.get('my.deep.nested.variable', coerce_type=str) == '1'

        assert p.get('my.deep.nested', coerce_type=dict) == {'variable': 1}

        assert p.get('hosts', coerce_type=list) == ['localhost', '127.0.0.1']

    def test__yaml_parser__scope(self):
        p = YamlParser('./tests/data/config.yml', scope='development')
        assert p.get('up.down.above') == [1, 2, 3]

    def test__yaml_parser__do_not_coerce_default_values(self):
        p = YamlParser('./tests/data/config.yml')
        o = object()
        assert p.get('does.not.exist', default=o, coerce_type=int) == o

    def test__yaml_parser__path_separator(self):
        p = YamlParser('./tests/data/config.yml', scope='development', path_separator='/')
        assert p.get('up/down/above') == [1, 2, 3]

    def test__get_client__raises(self):
        p = YamlParser('./tests/data/config.yml', scope='development', path_separator='/')
        with pytest.raises(NotImplementedError):
            p.get_client()



import io
import typing as t

from django_docker_helpers.config.backends.base import BaseParser
from django_docker_helpers.config.exceptions import KVStorageValueIsEmpty, KVStorageKeyDoestNotExist
from .yaml_parser import YamlParser


class ConsulParser(BaseParser):
    """
    Reads a whole config bundle from a consul kv key and provides the unified interface to access config options.

    It assumes that config in your storage can be parsed with any simple parser, like
    :class:`~django_docker_helpers.config.backends.YamlParser`.

    Compared to, e.g. :class:`~django_docker_helpers.config.backends.environment_parser.EnvironmentParser`
    it does not have scope support by design, since ``endpoint`` is a good enough scope by itself.

    Example:
    ::

        parser = ConsulParser('my/server/config.yml', host=CONSUL_HOST, port=CONSUL_PORT)
        parser.get('nested.a.b', coerce_type=int)
    """
    def __init__(self,
                 endpoint: str = 'service',
                 host: str = '127.0.0.1',
                 port: int = 8500,
                 scheme: str = 'http',
                 verify: bool = True,
                 cert: str = None,
                 kv_get_opts: t.Optional[t.Dict] = None,

                 path_separator: str = '.',
                 inner_parser_class: t.Optional[t.Type[BaseParser]] = YamlParser):
        """

        :param endpoint: specifies a key in consul kv storage, e.g. ``'services/mailer/config.yml'``
        :param host: consul host, default is ``'127.0.0.1'``
        :param port: consul port, default is ``8500``
        :param scheme: consul scheme, default is ``'http'``
        :param verify: verify certs, default is ``True``
        :param cert: path to certificate bundle
        :param kv_get_opts: read config bundle with optional arguments to ``client.kv.get()``
        :param path_separator: specifies which character separates nested variables, default is ``'.'``
        :param inner_parser_class: use the specified parser to read config from ``endpoint`` key
        """
        super().__init__(path_separator=path_separator)

        self.endpoint = endpoint

        self.client_options = {
            'host': host,
            'port': port,
            'scheme': scheme,
            'verify': verify,
            'cert': cert,
        }

        self.inner_parser_class = inner_parser_class
        self.kv_get_opts = kv_get_opts or {}

        self._inner_parser = None

    def __str__(self):
        return '<{0} {1[scheme]}://{1[host]}:{1[port]} scope={2}>'.format(
            self.__class__.__name__,
            self.client_options,
            self.scope
        )

    def get_client(self):
        # type: () -> consul.Consul
        import consul
        self._client = consul.Consul(**self.client_options)
        return self._client

    @property
    def inner_parser(self) -> BaseParser:
        """
        Prepares inner config parser for config stored at ``endpoint``.

        :return: an instance of :class:`~django_docker_helpers.config.backends.base.BaseParser`

        :raises config.exceptions.KVStorageKeyDoestNotExist: if specified ``endpoint`` does not exists

        :raises config.exceptions.KVStorageValueIsEmpty: if specified ``endpoint`` does not contain a config
        """
        if self._inner_parser is not None:
            return self._inner_parser

        __index, response_config = self.client.kv.get(self.endpoint, **self.kv_get_opts)
        if not response_config:
            raise KVStorageKeyDoestNotExist('Key does not exist: `{0}`'.format(self.endpoint))

        config = response_config['Value']
        if not config or config is self.sentinel:
            raise KVStorageValueIsEmpty('Read empty config by key `{0}`'.format(self.endpoint))

        config = config.decode()

        self._inner_parser = self.inner_parser_class(
            config=io.StringIO(config),
            path_separator=self.path_separator,
            scope=None
        )
        return self._inner_parser

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            **kwargs):
        """
        Reads a value of ``variable_path`` from consul kv storage.

        :param variable_path: a delimiter-separated path to a nested value
        :param default: default value if there's no object by specified path
        :param coerce_type: cast a type of a value to a specified one
        :param coercer: perform a type casting with specified callback
        :param kwargs: additional arguments inherited parser may need
        :return: value or default

        :raises config.exceptions.KVStorageKeyDoestNotExist: if specified ``endpoint`` does not exists

        :raises config.exceptions.KVStorageValueIsEmpty: if specified ``endpoint`` does not contain a config
        """

        return self.inner_parser.get(
            variable_path,
            default=default,
            coerce_type=coerce_type,
            coercer=coercer,
            **kwargs,
        )



import logging
import os
import typing as t

from django_docker_helpers.utils import coerce_str_to_bool


class BaseParser:
    """
    Base class to inherit from in custom parsers.
    """
    def __init__(self,
                 scope: t.Optional[str] = None,
                 config: t.Optional[str] = None,
                 nested_delimiter: str = '__',
                 path_separator: str = '.',
                 env: t.Optional[t.Dict[str, str]] = None):
        """
        All ``__init__`` arguments **MUST** be optional if you need
        :meth:`~django_docker_helpers.config.ConfigLoader.from_env`
        automatic parser initializer (it initializes parsers like ``parser_class(**parser_options)``).

        Since :class:`~django_docker_helpers.config.ConfigLoader` can initialize parsers from environment variables
        it's **recommended** to annotate argument types to provide a correct auto typecast.

        ``BaseParser`` creates a ``logger`` with name ``__class__.__name__``.

        ``BaseParser`` implements generic copying of following arguments without any backend-specific logic inside.

        :param scope: a global prefix to all underlying values
        :param config: optional config
        :param nested_delimiter: optional delimiter for environment backend
        :param path_separator: specifies which character separates nested variables, default is ``'.'``
        :param env: a dict with environment variables, default is ``os.environ``
        """

        self.config = config
        self.scope = scope
        self.nested_delimiter = nested_delimiter
        self.path_separator = path_separator
        self.env = env or os.environ
        self.sentinel = object()
        self._client = None

        self.logger = logging.getLogger(self.__class__.__name__)

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            **kwargs):
        """
        Inherited method should take all specified arguments.

        :param variable_path: a delimiter-separated path to a nested value
        :param default: default value if there's no object by specified path
        :param coerce_type: cast a type of a value to a specified one
        :param coercer: perform a type casting with specified callback
        :param kwargs: additional arguments inherited parser may need
        :return: value or default
        """
        raise NotImplementedError

    @staticmethod
    def coerce(val: t.Any,
               coerce_type: t.Optional[t.Type] = None,
               coercer: t.Optional[t.Callable] = None) -> t.Any:
        """
        Casts a type of ``val`` to ``coerce_type`` with ``coercer``.

        If ``coerce_type`` is bool and no ``coercer`` specified it uses
        :func:`~django_docker_helpers.utils.coerce_str_to_bool` by default.

        :param val: a value of any type
        :param coerce_type: any type
        :param coercer: provide a callback that takes ``val`` and returns a value with desired type
        :return: type casted value
        """
        if not coerce_type and not coercer:
            return val

        if coerce_type and type(val) is coerce_type:
            return val

        if coerce_type and coerce_type is bool and not coercer:
            coercer = coerce_str_to_bool

        if coercer is None:
            coercer = coerce_type

        return coercer(val)

    @property
    def client(self):
        """
        Helper property to lazy initialize and cache client. Runs
        :meth:`~django_docker_helpers.config.backends.base.BaseParser.get_client`.

        :return: an instance of backend-specific client
        """
        if self._client is not None:
            return self._client

        self._client = self.get_client()
        return self._client

    def get_client(self):
        """
        If your backend needs a client, inherit this method and use
        :meth:`~django_docker_helpers.config.backends.base.BaseParser.client` shortcut.

        :return: an instance of backend-specific client
        """
        raise NotImplementedError



import typing as t

from django_docker_helpers.config.backends.base import BaseParser
from django_docker_helpers.utils import dotkey


class YamlParser(BaseParser):
    """
    Provides a simple interface to read config options from Yaml.

    Example:
    ::

        p = YamlParser('./tests/data/config.yml', scope='development')
        assert p.get('up.down.above') == [1, 2, 3]
    """
    def __init__(self,
                 config: t.Optional[t.Union[str, t.TextIO]] = None,
                 path_separator: str = '.',
                 scope: t.Optional[str] = None):
        """
        :param config: a path to config file, or `TextIO` object
        :param path_separator: specifies which character separates nested variables, default is ``'.'``
        :param scope: a global namespace-like variable prefix

        :raises ValueError: if no config specified
        """
        super().__init__(scope=scope, config=config, path_separator=path_separator)

        self._data = None

        if not config:
            raise ValueError('Config should not be empty')

    def __str__(self):
        return '<{0} config="{1}" scope={2}>'.format(
            self.__class__.__name__,
            self.config if isinstance(self.config, str) else 'TextIO',
            self.scope,
        )

    @property
    def data(self):
        if self._data is not None:
            return self._data

        from yaml import load

        if isinstance(self.config, str):
            config = open(self.config)
        else:
            config = self.config

        self._data = load(config)
        return self._data

    def get_client(self):
        raise NotImplementedError

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            **kwargs):

        if self.scope:
            variable_path = '{0.scope}{0.path_separator}{1}'.format(self, variable_path)

        val = dotkey(self.data, variable_path, default=self.sentinel, separator=self.path_separator)

        if val is self.sentinel:
            return default

        return self.coerce(val, coerce_type=coerce_type, coercer=coercer)



from unittest.mock import patch

# noinspection PyPackageRequirements
import pytest

from django_docker_helpers.db import migrate, ensure_caches_alive, ensure_databases_alive
from django_docker_helpers.files import collect_static
from django_docker_helpers.management import create_admin

pytestmark = pytest.mark.management


# noinspection PyMethodMayBeStatic
class ManagementTest:
    @pytest.mark.django_db
    def test__migrate(self):
        with patch('django_docker_helpers.db.wf') as wf:
            assert migrate()
            assert any('[+]' in arg[0][0] for arg in wf.call_args_list)

    @pytest.mark.django_db
    def test__ensure_databases_alive(self):
        with patch('django_docker_helpers.db.wf') as wf:
            assert ensure_databases_alive(max_retries=1)
            assert any('[+]' in arg[0][0] for arg in wf.call_args_list)

    def test__ensure_caches_alive(self):
        with patch('django_docker_helpers.db.wf') as wf:
            assert ensure_caches_alive(max_retries=1)
            assert any('[+]' in arg[0][0] for arg in wf.call_args_list)

    def test__collect_static(self):
        with patch('django_docker_helpers.files.wf') as wf:
            assert collect_static()
            assert any('[+]' in arg[0][0] for arg in wf.call_args_list)

    @pytest.mark.django_db
    def test__create_admin(self):
        with patch('django_docker_helpers.management.wf') as wf:
            assert create_admin()
            assert any('[+]' in arg[0][0] for arg in wf.call_args_list)



import os
import typing as t

from django_docker_helpers.config.backends.base import BaseParser
from django_docker_helpers.utils import coerce_str_to_bool


class EnvironmentParser(BaseParser):
    """
    Provides a simple interface to read config options from environment variables.

    Example:
    ::

        from json import loads as json_load
        from yaml import load as yaml_load

        env = {
            'MY__VARIABLE': '33',
            'MY__NESTED__YAML__LIST__VARIABLE': '[33, 42]',
            'MY__NESTED__JSON__DICT__VARIABLE': '{"obj": true}',
        }

        parser = EnvironmentParser(env=env)
        assert p.get('my.variable') == '33'

        assert p.get('my.nested.yaml.list.variable',
                     coerce_type=list, coercer=yaml_load) == [33, 42]
        assert p.get('my.nested.json.dict.variable',
                     coerce_type=dict, coercer=json_load) == {'obj': True}

        parser = EnvironmentParser(env=env, scope='my.nested')
        assert parser.get('yaml.list.variable',
                          coerce_type=list, coercer=yaml_load) == [33, 42]
    """
    def __init__(self,
                 scope: t.Optional[str] = None,
                 config: t.Optional[str] = None,
                 nested_delimiter: str = '__',
                 path_separator: str = '.',
                 env: t.Optional[t.Dict[str, str]] = None):
        """
        :param scope: a global namespace-like variable prefix
        :param config: not used
        :param nested_delimiter: replace ``path_separator`` with an appropriate environment variable delimiter,
         default is ``__``
        :param path_separator: specifies which character separates nested variables, default is ``'.'``
        :param env: a dict with environment variables, default is ``os.environ``
        """
        env = env or os.environ
        super().__init__(
            scope=scope, config=config,
            path_separator=path_separator, nested_delimiter=nested_delimiter,
            env=env,
        )
        # config is not used in EnvironmentParser
        self.config = None
        self.scope = (scope or '').upper()

    def __str__(self):
        return '<{0} scope={1}>'.format(
            self.__class__.__name__,
            self.scope
        )

    def get_client(self):
        raise NotImplementedError

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            **kwargs):
        """
        Reads a value of ``variable_path`` from environment.

        If ``coerce_type`` is ``bool`` and no ``coercer`` specified, ``coerces`` forced to be
        :func:`~django_docker_helpers.utils.coerce_str_to_bool`

        :param variable_path: a delimiter-separated path to a nested value
        :param default: default value if there's no object by specified path
        :param coerce_type: cast a type of a value to a specified one
        :param coercer: perform a type casting with specified callback
        :param kwargs: additional arguments inherited parser may need
        :return: value or default
        """

        var_name = self.get_env_var_name(variable_path)
        val = self.env.get(var_name, self.sentinel)
        if val is self.sentinel:
            return default

        # coerce to bool with default env coercer if no coercer specified
        if coerce_type and coerce_type is bool and not coercer:
            coercer = coerce_str_to_bool

        return self.coerce(val, coerce_type=coerce_type, coercer=coercer)

    def get_env_var_name(self, variable_path: str) -> str:
        return self.nested_delimiter.join(
            filter(
                None,
                (self.scope or '').split(self.path_separator) +
                variable_path.upper().split(self.path_separator)
            )
        )



import typing as t

from yaml import load as yaml_load

from django_docker_helpers.config.backends.base import BaseParser


class MPTRedisParser(BaseParser):
    """
    Materialized Path Tree Redis Parser.

    Compared to, e.g. :class:`~django_docker_helpers.config.backends.redis_parser.RedisParser`
    it does not load a whole config file from a single key, but reads every config option
    from a corresponding variable path.

    Example:
    ::

        parser = MPTRedisParser(host=REDIS_HOST, port=REDIS_PORT)
        parser.get('nested.a.b')
        parser.get('debug')

    If you want to store your config with separated key paths take
    :func:`~django_docker_helpers.utils.mp_serialize_dict` helper to materialize your dict.
    """

    def __init__(self,
                 scope: t.Optional[str] = None,
                 host: str = '127.0.0.1',
                 port: int = 6379,
                 db: int = 0,
                 path_separator: str = '.',
                 key_prefix: str = '',
                 object_deserialize_prefix: str = '::YAML::\n',
                 object_deserialize: t.Optional[t.Callable] = yaml_load,
                 **redis_options):
        """

        :param scope: a global namespace-like variable prefix
        :param host: redis host, default is ``'127.0.0.1'``
        :param port: redis port, default id ``6379``
        :param db: redis database, default is ``0``
        :param path_separator: specifies which character separates nested variables, default is ``'.'``
        :param key_prefix: prefix all keys with specified one
        :param object_deserialize_prefix: if object has a specified prefix, it's deserialized with
         ``object_deserialize``
        :param object_deserialize: deserializer for complex variables
        :param redis_options: additional options for ``redis.Redis`` client
        """

        super().__init__(scope=scope, path_separator=path_separator)
        self.object_serialize_prefix = object_deserialize_prefix.encode()
        self.object_deserialize = object_deserialize
        self.key_prefix = key_prefix

        self.client_options = {
            'host': host,
            'port': port,
            'db': db,
        }
        self.client_options.update(**redis_options)

    def __str__(self):
        return '<{0} {1[host]}:{1[port]} db={1[db]} scope={2}>'.format(
            self.__class__.__name__,
            self.client_options,
            self.scope,
        )

    def get_client(self):
        # type: () -> redis.Redis
        import redis
        self._client = redis.Redis(**self.client_options)
        return self._client

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            **kwargs):
        """
        :param variable_path: a delimiter-separated path to a nested value
        :param default: default value if there's no object by specified path
        :param coerce_type: cast a type of a value to a specified one
        :param coercer: perform a type casting with specified callback
        :param kwargs: additional arguments inherited parser may need
        :return: value or default
        """

        if self.scope:
            variable_path = '{0.scope}{0.path_separator}{1}'.format(self, variable_path)

        if self.key_prefix:
            variable_path = '{0.key_prefix}:{1}'.format(self, variable_path)

        val = self.client.get(variable_path)

        if val is None:
            return default

        if val.startswith(self.object_serialize_prefix):
            # since complex data types are yaml-serialized there's no need to coerce anything
            _val = val[len(self.object_serialize_prefix):]
            bundle = self.object_deserialize(_val)
            if bundle == '':  # check for reinforced empty flag
                return self.coerce(bundle, coerce_type=coerce_type, coercer=coercer)
            return bundle

        if isinstance(val, bytes):
            val = val.decode()

        return self.coerce(val, coerce_type=coerce_type, coercer=coercer)



from django_docker_helpers.utils import wf, run_env_once


@run_env_once
def collect_static() -> bool:
    """
    Runs Django ``collectstatic`` command in silent mode.

    :return: always ``True``
    """
    from django.core.management import execute_from_command_line
    # from django.conf import settings
    # if not os.listdir(settings.STATIC_ROOT):
    wf('Collecting static files... ', False)
    execute_from_command_line(['./manage.py', 'collectstatic', '-c', '--noinput', '-v0'])
    wf('[+]\n')
    return True



import io
import typing as t

from django_docker_helpers.config.backends.base import BaseParser
from django_docker_helpers.config.exceptions import KVStorageValueIsEmpty
from .yaml_parser import YamlParser


class RedisParser(BaseParser):
    """
    Reads a whole config bundle from a redis key and provides the unified interface to access config options.

    It assumes that config in your storage can be parsed with any simple parser, like
    :class:`~django_docker_helpers.config.backends.YamlParser`.

    Compared to, e.g. :class:`~django_docker_helpers.config.backends.environment_parser.EnvironmentParser`
    it does not have scope support by design, since ``endpoint`` is a good enough scope by itself.

    Example:
    ::

        parser = RedisParser('my/server/config.yml', host=REDIS_HOST, port=REDIS_PORT)
        parser.get('nested.a.b', coerce_type=int)

    """
    def __init__(self,
                 endpoint: str = 'service',
                 host: str = '127.0.0.1',
                 port: int = 6379,
                 db: int = 0,
                 path_separator: str = '.',
                 inner_parser_class: t.Optional[t.Type[BaseParser]] = YamlParser,
                 **redis_options):
        """

        :param endpoint: specifies a redis key with serialized config, e.g. ``'services/mailer/config.yml'``
        :param host: redis host, default is ``'127.0.0.1'``
        :param port: redis port, default id ``6379``
        :param db: redis database, default is ``0``
        :param path_separator: specifies which character separates nested variables, default is ``'.'``
        :param inner_parser_class: use the specified parser to read config from ``endpoint`` key
        :param redis_options: additional options for ``redis.Redis`` client
        """

        super().__init__(path_separator=path_separator)
        self.inner_parser_class = inner_parser_class
        self.endpoint = endpoint

        self.client_options = {
            'host': host,
            'port': port,
            'db': db,
        }
        self.client_options.update(**redis_options)
        self._inner_parser = None

    def __str__(self):
        return '<{0} {1[host]}:{1[port]} db={1[db]} scope={2}>'.format(
            self.__class__.__name__,
            self.client_options,
            self.scope,
        )

    def get_client(self):
        # type: () -> redis.Redis
        import redis
        self._client = redis.Redis(**self.client_options)
        return self._client

    @property
    def inner_parser(self) -> BaseParser:
        """
        Prepares inner config parser for config stored at ``endpoint``.

        :return: an instance of :class:`~django_docker_helpers.config.backends.base.BaseParser`

        :raises config.exceptions.KVStorageValueIsEmpty: if specified ``endpoint`` does not contain a config
        """
        if self._inner_parser is not None:
            return self._inner_parser

        config = self.client.get(self.endpoint)
        if not config:
            raise KVStorageValueIsEmpty('Key `{0}` does not exist or value is empty'.format(self.endpoint))

        config = config.decode()

        self._inner_parser = self.inner_parser_class(
            config=io.StringIO(config),
            path_separator=self.path_separator,
            scope=None
        )
        return self._inner_parser

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            **kwargs):

        """
        Reads a value of ``variable_path`` from redis storage.

        :param variable_path: a delimiter-separated path to a nested value
        :param default: default value if there's no object by specified path
        :param coerce_type: cast a type of a value to a specified one
        :param coercer: perform a type casting with specified callback
        :param kwargs: additional arguments inherited parser may need
        :return: value or default

        :raises config.exceptions.KVStorageValueIsEmpty: if specified ``endpoint`` does not contain a config
        """

        return self.inner_parser.get(
            variable_path,
            default=default,
            coerce_type=coerce_type,
            coercer=coercer,
            **kwargs,
        )



from io import StringIO
import os
from unittest import mock

# noinspection PyPackageRequirements
import pytest

from django.core.management import call_command

pytestmark = [pytest.mark.cli, pytest.mark.django]

ENDPOINT = '__DJANGO_DOCKER_HELPERS__:uwsgi-config.yml'

EXPECTED_FILE_CONTENT = """[uwsgi]
socket = 0.0.0.0:80
static-map = /test/=/application/test
static-map = /another-test/=/application/another-test
"""


@pytest.fixture
def store_redis_config():
    from redis import Redis
    from yaml import dump

    cfg = dump({
        'uwsgi': {
            'socket': '0.0.0.0:80',
            'static-map': [
                '/test/=/application/test',
                '/another-test/=/application/another-test'
            ]
        }
    }).encode()

    client = Redis(
        host=os.getenv('REDIS_HOST', '127.0.0.1'),
        port=os.getenv('REDIS_PORT', 6379)
    )
    client.set(ENDPOINT, cfg)
    yield
    client.delete(ENDPOINT)


# noinspection PyMethodMayBeStatic
@pytest.mark.django_db
class DjangoCliTest:
    def test_command_output(self, store_redis_config):
        with mock.patch(
                'django_docker_helpers.cli.django.management.commands.run_configured_uwsgi.os.execvp'
        ) as mocked__os_execvp:
            os.environ.setdefault('CONFIG__PARSERS', 'EnvironmentParser,RedisParser')
            os.environ.setdefault('REDISPARSER__ENDPOINT', ENDPOINT)

            out = StringIO()
            call_command('run_configured_uwsgi', '--print', '-f', '/tmp/__uwsgi.ini', stdout=out)

            # check output (lazy checking =)
            assert EXPECTED_FILE_CONTENT in out.getvalue()

            # check config file
            assert os.path.exists('/tmp/__uwsgi.ini')
            assert os.path.isfile('/tmp/__uwsgi.ini')
            with open('/tmp/__uwsgi.ini', 'r') as f:
                assert EXPECTED_FILE_CONTENT == f.read()

            # check os.execvp arguments
            mocked__os_execvp.assert_called_once()
            mocked__os_execvp.assert_called_once_with('uwsgi', ('--ini', '/tmp/__uwsgi.ini'))

            # NOTE: it's necessary to make environ clean
            os.environ.clear()



import os

DEBUG = True

SECRET_KEY = 'lol'

STATIC_URL = '/static/'
STATIC_ROOT = './static/'

INSTALLED_APPS = [
    'django.contrib.staticfiles',
    'django.contrib.auth',
    'django.contrib.contenttypes',

    'django_docker_helpers.cli.django',

    'tests.test_app',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'tests/test.sqlite',
    }
}

MEMCACHED_HOST = os.getenv('MEMCACHED_HOST', '127.0.0.1')
MEMCACHED_PORT = os.getenv('MEMCACHED_HOST', '11211')

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
        'LOCATION': '{0}:{1}'.format(MEMCACHED_HOST, MEMCACHED_PORT),
    }
}

CONFIG = {
    'superuser': {
        'username': 'admin',
        'email': 'admin@example.com',
        'password': 'r00t',
    }
}



import os

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
from yaml import dump as yaml_dump

from django_docker_helpers.config import exceptions
from django_docker_helpers.config.backends.redis_parser import RedisParser

pytestmark = [pytest.mark.backend, pytest.mark.redis]

REDIS_HOST = os.getenv('REDIS_HOST', '127.0.0.1')
REDIS_PORT = os.getenv('REDIS_PORT', 6379)

SAMPLE = {
    'bool_flag': '',  # flag
    'unicode': 'вася',
    'none_value': None,
    'debug': True,
    'mixed': ['ascii', 'юникод', 1, {'d': 1}, {'b': 2}],
    'nested': {
        'a': {
            'b': 2
        }
    }
}


@pytest.fixture
def redis_parser():
    return RedisParser('my/server/config.yml', host=REDIS_HOST, port=REDIS_PORT)


@pytest.fixture
def store_redis_config():
    import redis

    c = redis.Redis(host=REDIS_HOST, port=REDIS_PORT)
    data = yaml_dump(SAMPLE, allow_unicode=True).encode()
    c.set('my/server/config.yml', data)
    c.set('my/server/empty.yml', '')
    return c


# noinspection PyMethodMayBeStatic,PyShadowingNames,PyUnusedLocal
class RedisBackendTest:
    def test__redis_parser__client(self, redis_parser: RedisParser):
        assert redis_parser.client

    def test__redis_parser__inner_parser(self, store_redis_config, redis_parser: RedisParser):
        assert redis_parser.client
        assert redis_parser.inner_parser

    def test__redis_parser__inner_parser__exceptions(self, store_redis_config):
        with pytest.raises(exceptions.KVStorageValueIsEmpty):
            c = RedisParser('nothing/here', host=REDIS_HOST, port=REDIS_PORT)
            assert c.inner_parser

        with pytest.raises(exceptions.KVStorageValueIsEmpty):
            c = RedisParser('my/server/empty.yml', host=REDIS_HOST, port=REDIS_PORT)
            assert c.inner_parser

    def test__redis_parser__configs_equal(self, store_redis_config, redis_parser):
        assert redis_parser.inner_parser
        assert redis_parser.inner_parser.data == SAMPLE



import importlib
import os
import sys
import typing as t
from decimal import Decimal
from functools import wraps

from dpath.util import get
from yaml import dump as dump_yaml

ENV_STR_BOOL_COERCE_MAP = {
    '': True,  # Flag is set

    0: False,
    '0': False,
    'false': False,
    'off': False,

    1: True,
    '1': True,
    'true': True,
    'on': True,
}

SHRED_DATA_FIELD_NAMES = (
    'password',
    'secret',
    'pass',
    'pwd',
    'key',
    'token',
    'auth',
    'cred',
)


def shred(key_name: str,
          value: t.Any,
          field_names: t.Iterable[str] = SHRED_DATA_FIELD_NAMES) -> t.Union[t.Any, str]:
    """
    Replaces sensitive data in ``value`` with ``*`` if ``key_name`` contains something that looks like a secret.

    :param field_names: a list of key names that can possibly contain sensitive data
    :param key_name: a key name to check
    :param value: a value to mask
    :return: an unchanged value if nothing to hide, ``'*' * len(str(value))`` otherwise
    """
    key_name = key_name.lower()
    need_shred = False
    for data_field_name in field_names:
        if data_field_name in key_name:
            need_shred = True
            break

    if not need_shred:
        return value

    return '*' * len(str(value))


def shred_deep(value, field_names: t.Iterable[str] = SHRED_DATA_FIELD_NAMES):
    if isinstance(value, (int, str)):
        return value

    if isinstance(value, dict):
        _value = {}
        for k, v in value.items():
            if isinstance(v, (int, str)):
                _value[k] = shred(k, v, field_names=field_names)
            else:
                _value[k] = shred_deep(v, field_names=field_names)
        return _value

    _value = [shred_deep(v, field_names=field_names) for v in value]
    return type(value)(_value)


def import_from(module: str, name: str):
    return getattr(
        importlib.import_module(module, [name]),
        name
    )


def dot_path(obj: t.Union[t.Dict, object],
             path: str,
             default: t.Any = None,
             separator: str = '.'):
    """
    Provides an access to elements of a mixed dict/object type by a delimiter-separated path.
    ::

        class O1:
            my_dict = {'a': {'b': 1}}

        class O2:
            def __init__(self):
                self.nested = O1()

        class O3:
            final = O2()

        o = O3()
        assert utils.dot_path(o, 'final.nested.my_dict.a.b') == 1

    .. testoutput::

        True

    :param obj: object or dict
    :param path: path to value
    :param default: default value if chain resolve failed
    :param separator: ``.`` by default
    :return: value or default
    """
    path_items = path.split(separator)
    val = obj
    sentinel = object()
    for item in path_items:
        if isinstance(val, dict):
            val = val.get(item, sentinel)
            if val is sentinel:
                return default
        else:
            val = getattr(val, item, sentinel)
            if val is sentinel:
                return default
    return val


def dotkey(obj: dict, path: str, default=None, separator='.'):
    """
    Provides an interface to traverse nested dict values by dot-separated paths. Wrapper for ``dpath.util.get``.

    :param obj: dict like ``{'some': {'value': 3}}``
    :param path: ``'some.value'``
    :param separator: ``'.'`` or ``'/'`` or whatever
    :param default: default for KeyError
    :return: dict value or default value
    """
    try:
        return get(obj, path, separator=separator)
    except KeyError:
        return default


def _materialize_dict(bundle: dict, separator: str = '.') -> t.Generator[t.Tuple[str, t.Any], None, None]:
    """
    Traverses and transforms a given dict ``bundle`` into tuples of ``(key_path, value)``.

    :param bundle: a dict to traverse
    :param separator: build paths with a given separator
    :return: a generator of tuples ``(materialized_path, value)``

    Example:
    >>> list(_materialize_dict({'test': {'path': 1}, 'key': 'val'}, '.'))
    >>> [('key', 'val'), ('test.path', 1)]
    """
    for path_prefix, v in bundle.items():
        if not isinstance(v, dict):
            yield str(path_prefix), v
            continue

        for nested_path, nested_val in _materialize_dict(v, separator=separator):
            yield '{0}{1}{2}'.format(path_prefix, separator, nested_path), nested_val


def materialize_dict(bundle: dict, separator: str = '.') -> t.List[t.Tuple[str, t.Any]]:
    """
    Transforms a given ``bundle`` into a *sorted* list of tuples with materialized value paths and values:
    ``('path.to.value', <value>)``. Output is ordered by depth: the deepest element first.

    :param bundle: a dict to materialize
    :param separator: build paths with a given separator
    :return: a depth descending and alphabetically ascending sorted list (-deep, asc), the longest first

    ::

        sample = {
            'a': 1,
            'aa': 1,
            'b': {
                'c': 1,
                'b': 1,
                'a': 1,
                'aa': 1,
                'aaa': {
                    'a': 1
                }
            }
        }
        materialize_dict(sample, '/')
        [
            ('b/aaa/a', 1),
            ('b/a', 1),
            ('b/aa', 1),
            ('b/b', 1),
            ('b/c', 1),
            ('a', 1),
            ('aa', 1)
        ]
    """

    def _matkeysort(tup: t.Tuple[str, t.Any]):
        return len(tup[0].split(separator))

    s1 = sorted(_materialize_dict(bundle, separator=separator), key=lambda x: x[0])
    return sorted(s1, key=_matkeysort, reverse=True)


def mp_serialize_dict(
        bundle: dict,
        separator: str = '.',
        serialize: t.Optional[t.Callable] = dump_yaml,
        value_prefix: str = '::YAML::\n') -> t.List[t.Tuple[str, bytes]]:
    """
    Transforms a given ``bundle`` into a *sorted* list of tuples with materialized value paths and values:
    ``('path.to.value', b'<some>')``. If the ``<some>`` value is not an instance of a basic type, it's serialized
    with ``serialize`` callback. If this value is an empty string, it's serialized anyway to enforce correct
    type if storage backend does not support saving empty strings.

    :param bundle: a dict to materialize
    :param separator: build paths with a given separator
    :param serialize: a method to serialize non-basic types, default is ``yaml.dump``
    :param value_prefix: a prefix for non-basic serialized types
    :return: a list of tuples ``(mat_path, b'value')``

    ::

        sample = {
            'bool_flag': '',  # flag
            'unicode': 'вася',
            'none_value': None,
            'debug': True,
            'mixed': ['ascii', 'юникод', 1, {'d': 1}, {'b': 2}],
            'nested': {
                'a': {
                    'b': 2,
                    'c': b'bytes',
                }
            }
        }

        result = mp_serialize_dict(sample, separator='/')
        assert result == [
            ('nested/a/b', b'2'),
            ('nested/a/c', b'bytes'),
            ('bool_flag', b"::YAML::\\n''\\n"),
            ('debug', b'true'),
            ('mixed', b'::YAML::\\n- ascii\\n- '
                      b'"\\\\u044E\\\\u043D\\\\u0438\\\\u043A\\\\u043E\\\\u0434"\\n- 1\\n- '
                      b'{d: 1}\\n- {b: 2}\\n'),
            ('none_value', None),
            ('unicode', b'\\xd0\\xb2\\xd0\\xb0\\xd1\\x81\\xd1\\x8f')
        ]
    """

    md = materialize_dict(bundle, separator=separator)
    res = []
    for path, value in md:
        # have to serialize values (value should be None or a string / binary data)
        if value is None:
            pass
        elif isinstance(value, str) and value != '':
            # check for value != '' used to armor empty string with forced serialization
            # since it can be not recognized by a storage backend
            pass
        elif isinstance(value, bytes):
            pass
        elif isinstance(value, bool):
            value = str(value).lower()
        elif isinstance(value, (int, float, Decimal)):
            value = str(value)
        else:
            value = (value_prefix + serialize(value))

        if isinstance(value, str):
            value = value.encode()

        res.append((path, value))

    return res


def wf(raw_str: str,
       flush: bool = True,
       prevent_completion_polluting: bool = True,
       stream: t.TextIO = sys.stdout):
    """
    Writes a given ``raw_str`` into a ``stream``. Ignores output if ``prevent_completion_polluting`` is set and there's
    no extra ``sys.argv`` arguments present (a bash completion issue).

    :param raw_str: a raw string to print
    :param flush: execute ``flush()``
    :param prevent_completion_polluting: don't write anything if ``len(sys.argv) <= 1``
    :param stream: ``sys.stdout`` by default
    :return: None
    """
    if prevent_completion_polluting and len(sys.argv) <= 1:
        return

    stream.write(raw_str)
    flush and hasattr(stream, 'flush') and stream.flush()


def coerce_str_to_bool(val: t.Union[str, int, bool, None], strict: bool = False) -> bool:
    """
    Converts a given string ``val`` into a boolean.

    :param val: any string representation of boolean
    :param strict: raise ``ValueError`` if ``val`` does not look like a boolean-like object
    :return: ``True`` if ``val`` is thruthy, ``False`` otherwise.

    :raises ValueError: if ``strict`` specified and ``val`` got anything except
     ``['', 0, 1, true, false, on, off, True, False]``
    """
    if isinstance(val, str):
        val = val.lower()

    flag = ENV_STR_BOOL_COERCE_MAP.get(val, None)

    if flag is not None:
        return flag

    if strict:
        raise ValueError('Unsupported value for boolean flag: `%s`' % val)

    return bool(val)


def env_bool_flag(flag_name: str, strict: bool = False, env: t.Optional[t.Dict[str, str]] = None) -> bool:
    """
    Converts an environment variable into a boolean. Empty string (presence in env) is treated as ``True``.

    :param flag_name: an environment variable name
    :param strict: raise ``ValueError`` if a ``flag_name`` value connot be coerced into a boolean in obvious way
    :param env: a dict with environment variables, default is ``os.environ``
    :return: ``True`` if ``flag_name`` is thruthy, ``False`` otherwise.

    :raises ValueError: if ``strict`` specified and ``val`` got anything except ``['', 0, 1, true, false, True, False]``
    """
    env = env or os.environ
    sentinel = object()
    val = env.get(flag_name, sentinel)

    if val is sentinel:
        return False

    return coerce_str_to_bool(val, strict=strict)


def run_env_once(f: t.Callable) -> t.Callable:
    """
    A decorator to prevent ``manage.py`` from running code twice for everything.
    (https://stackoverflow.com/questions/16546652/why-does-django-run-everything-twice)

    :param f: function or method to decorate
    :return: callable
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        has_run = os.environ.get(wrapper.__name__)
        if has_run == '1':
            return
        result = f(*args, **kwargs)
        os.environ[wrapper.__name__] = '1'
        return result

    return wrapper


def is_dockerized(flag_name: str = 'DOCKERIZED', strict: bool = False):
    """
    Reads env ``DOCKERIZED`` variable as a boolean.

    :param flag_name: environment variable name
    :param strict: raise a ``ValueError`` if variable does not look like a normal boolean
    :return: ``True`` if has truthy ``DOCKERIZED`` env, ``False`` otherwise
    """
    return env_bool_flag(flag_name, strict=strict)


def is_production(flag_name: str = 'PRODUCTION', strict: bool = False):
    """
    Reads env ``PRODUCTION`` variable as a boolean.

    :param flag_name: environment variable name
    :param strict: raise a ``ValueError`` if variable does not look like a normal boolean
    :return: ``True`` if has truthy ``PRODUCTION`` env, ``False`` otherwise
    """
    return env_bool_flag(flag_name, strict=strict)



from setuptools import setup, find_packages

from django_docker_helpers import __version__

try:
    import pypandoc

    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = ''

setup(
    name='django-docker-helpers',
    version=__version__,
    packages=find_packages(),
    url='https://github.com/night-crawler/django-docker-helpers',
    license='MIT',
    author='night-crawler',
    author_email='lilo.panic@gmail.com',
    description='Django Docker helpers',
    long_description=long_description,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=['dpath', 'pyaml', 'gunicorn', 'django', 'terminaltables']
)



import typing as t

from yaml import load as yaml_load

from django_docker_helpers.config.backends.base import BaseParser


class MPTConsulParser(BaseParser):
    """
    Materialized Path Tree Consul Parser.

    Compared to, e.g. :class:`~django_docker_helpers.config.backends.consul_parser.ConsulParser`
    it does not load a whole config file from a single key, but reads every config option
    from a corresponding variable path.

    Example:
    ::

        parser = MPTConsulParser(host=CONSUL_HOST, port=CONSUL_PORT, path_separator='.')
        parser.get('nested.a.b')

    If you want to store your config with separated key paths take
    :func:`~django_docker_helpers.utils.mp_serialize_dict` helper to materialize your dict.
    """
    def __init__(self,
                 scope: t.Optional[str] = None,
                 host: str = '127.0.0.1',
                 port: int = 8500,
                 scheme: str = 'http',
                 verify: bool = True,
                 cert=None,
                 path_separator: str = '.',
                 consul_path_separator: str = '/',
                 object_deserialize_prefix: str = '::YAML::\n',
                 object_deserialize: t.Optional[t.Callable] = yaml_load):
        """
        :param scope: a global namespace-like variable prefix
        :param host: consul host, default is ``'127.0.0.1'``
        :param port: consul port, default is ``8500``
        :param scheme: consul scheme, default is ``'http'``
        :param verify: verify certs, default is ``True``
        :param cert: path to certificate bundle
        :param path_separator: specifies which character separates nested variables,
         default is ``'.'``
        :param consul_path_separator: specifies which character separates nested variables in consul kv storage,
         default is ``'/'``
        :param object_deserialize_prefix: if object has a specified prefix, it's deserialized with
         ``object_deserialize``
        :param object_deserialize: deserializer for complex variables
        """
        super().__init__(scope=scope, path_separator=path_separator)
        self.object_serialize_prefix = object_deserialize_prefix.encode()
        self.object_deserialize = object_deserialize
        self.consul_path_separator = consul_path_separator

        self.client_options = {
            'host': host,
            'port': port,
            'scheme': scheme,
            'verify': verify,
            'cert': cert,
        }

    def __str__(self):
        return '<{0} {1[scheme]}://{1[host]}:{1[port]} scope={2}>'.format(
            self.__class__.__name__,
            self.client_options,
            self.scope,
        )

    def get_client(self):
        # type: () -> consul.Consul
        import consul
        self._client = consul.Consul(**self.client_options)
        return self._client

    def get(self,
            variable_path: str,
            default: t.Optional[t.Any] = None,
            coerce_type: t.Optional[t.Type] = None,
            coercer: t.Optional[t.Callable] = None,
            **kwargs):
        """
        :param variable_path: a delimiter-separated path to a nested value
        :param default: default value if there's no object by specified path
        :param coerce_type: cast a type of a value to a specified one
        :param coercer: perform a type casting with specified callback
        :param kwargs: additional arguments inherited parser may need
        :return: value or default
        """

        if self.path_separator != self.consul_path_separator:
            variable_path = variable_path.replace(self.path_separator, self.consul_path_separator)

        if self.scope:
            _scope = self.consul_path_separator.join(self.scope.split(self.path_separator))
            variable_path = '{0}/{1}'.format(_scope, variable_path)

        index, data = self.client.kv.get(variable_path, **kwargs)

        if data is None:
            return default

        val = data['Value']
        if val is None:
            # None is present and it is a valid value
            return val

        if val.startswith(self.object_serialize_prefix):
            # since complex data types are yaml-serialized there's no need to coerce anything
            _val = val[len(self.object_serialize_prefix):]
            bundle = self.object_deserialize(_val)
            if bundle == '':  # check for reinforced empty flag
                return self.coerce(bundle, coerce_type=coerce_type, coercer=coercer)
            return bundle

        if isinstance(val, bytes):
            val = val.decode()

        return self.coerce(val, coerce_type=coerce_type, coercer=coercer)



from django.core.handlers.wsgi import WSGIHandler
from django.core.management import execute_from_command_line

from django_docker_helpers.utils import wf, run_env_once, dot_path


@run_env_once
def create_admin(user_config_path: str = 'CONFIG.superuser') -> bool:
    """
    Creates a superuser from a specified dict/object bundle located at ``user_config_path``.
    Skips if the specified object contains no email or no username.
    If a user with the specified username already exists and has no usable password it updates user's password with
    a specified one.

    ``user_config_path`` can accept any path to a deep nested object, like dict of dicts,
    object of dicts of objects, and so on. Let's assume you have this weird config in your ``settings.py``:
    ::

        class MyConfigObject:
            my_var = {
                'user': {
                    'username': 'user',
                    'password': 'qwe',
                    'email': 'no@example.com',
                }
            }
        local_config = MyConfigObject()

    To access the ``'user'`` bundle you have to specify: ``local_config.my_var.user``.

    :param user_config_path: dot-separated path to object or dict, default is ``'CONFIG.superuser'``
    :return: ``True`` if user has been created, ``False`` otherwise
    """
    from django.conf import settings
    wf('Creating superuser... ', False)
    username, email, password = [
        dot_path(settings, '{0}.{1}'.format(user_config_path, 'username')),
        dot_path(settings, '{0}.{1}'.format(user_config_path, 'email')),
        dot_path(settings, '{0}.{1}'.format(user_config_path, 'password')),
    ]
    if not all([username, email]):
        wf('[SKIP: username and email should not be empty]\n')
        return False

    from django.db import IntegrityError
    try:
        execute_from_command_line([
            './manage.py', 'createsuperuser',
            '--username', username,
            '--email', email,
            '--noinput'
        ])
    except IntegrityError:
        pass

    if password:
        # after `execute_from_command_line` models are loaded
        from django.contrib.auth.models import User
        user = User.objects.get(username=username)

        # do not change password if it was set before
        if not user.has_usable_password():
            user.set_password(password)
            user.save()
    else:
        wf('[SKIP update password: password is empty]\n')

    wf('[+]\n')
    return True


def run_gunicorn(application: WSGIHandler, gunicorn_module_name: str = 'gunicorn_prod'):
    """
    Runs gunicorn with a specified config.

    :param application: Django uwsgi application
    :param gunicorn_module_name: gunicorn settings module name
    :return: ``Application().run()``
    """
    from gunicorn.app.base import Application

    class DjangoApplication(Application):
        def init(self, parser, opts, args):
            cfg = self.get_config_from_module_name(gunicorn_module_name)
            clean_cfg = {}
            for k, v in cfg.items():
                # Ignore unknown names
                if k not in self.cfg.settings:
                    continue
                clean_cfg[k.lower()] = v
            return clean_cfg

        def load(self) -> WSGIHandler:
            return application

    return DjangoApplication().run()



import os

# noinspection PyPackageRequirements
import pytest

from django_docker_helpers.config import ConfigLoader
from django_docker_helpers.config import exceptions
from django_docker_helpers.config.backends import *
from django_docker_helpers.utils import mp_serialize_dict

pytestmark = pytest.mark.config_loader

REDIS_HOST = os.getenv('REDIS_HOST', '127.0.0.1')
REDIS_PORT = os.getenv('REDIS_PORT', 6379)

CONSUL_HOST = os.getenv('CONSUL_HOST', '127.0.0.1')
CONSUL_PORT = os.getenv('CONSUL_PORT', 8500)


@pytest.fixture
def store_mpt_consul_config():
    import consul
    sample = {
        'project': {
            'variable': 2
        }
    }
    c = consul.Consul(host=CONSUL_HOST, port=CONSUL_PORT)
    for path, value in mp_serialize_dict(sample, separator='/'):
        c.kv.put(path, value)
    return c


@pytest.fixture
def store_consul_config():
    import consul
    from yaml import dump
    sample = {
        'some': {
            'variable': 2,
        }
    }
    c = consul.Consul(host=CONSUL_HOST, port=CONSUL_PORT)
    data = dump(sample).encode()
    c.kv.put('my/service/config.yml', data)
    return c


@pytest.fixture
def store_mpt_redis_config():
    import redis
    sample = {
        'project': {
            'i': {
                'am': {
                    'redis': True
                }
            }
        }
    }
    c = redis.Redis(host=REDIS_HOST, port=REDIS_PORT)
    for path, value in mp_serialize_dict(sample, separator='.'):
        c.set(path, value)
        c.set('my-prefix:%s' % path, value)

    return c


@pytest.fixture
def store_redis_config():
    from redis import Redis
    from yaml import dump

    sample = {
        'some': {
            'variable': 44,
            'brutal': 666
        }
    }
    c = Redis(host=REDIS_HOST, port=REDIS_PORT)
    data = dump(sample).encode()
    c.set('my/conf/service/config.yml', data)
    return c


@pytest.fixture
def loader():
    env = {
        'PROJECT__DEBUG': 'false'
    }
    parsers = [
        EnvironmentParser(scope='project', env=env),

        MPTConsulParser(host=CONSUL_HOST, port=CONSUL_PORT, scope='project'),
        ConsulParser('my/service/config.yml', host=CONSUL_HOST, port=CONSUL_PORT),

        MPTRedisParser(host=REDIS_HOST, port=REDIS_PORT, scope='project'),
        RedisParser('my/conf/service/config.yml', host=REDIS_HOST, port=REDIS_PORT),

        YamlParser(config='./tests/data/config.yml', scope='project'),
    ]

    return ConfigLoader(parsers=parsers)


# noinspection PyMethodMayBeStatic,PyShadowingNames,PyUnusedLocal
class ConfigLoaderTest:
    def test__priority(self,
                       loader: ConfigLoader,
                       store_mpt_consul_config,
                       store_mpt_redis_config,
                       store_consul_config,
                       store_redis_config):
        assert loader.get('debug') == 'false', 'Ensure value is taken from env'
        assert loader.get('debug', coerce_type=bool) is False, 'Ensure value is coercing properly for env'

        assert loader.get('variable', coerce_type=int) == 2, 'Ensure consul MPT backend attached'
        assert loader.get('i.am.redis', coerce_type=bool) is True, 'Ensure redis MPT backend attached'

        assert loader.get('some.variable', coerce_type=int) == 2, 'Ensure consul backend attached'
        assert loader.get('some.brutal', coerce_type=int) == 666, 'Ensure redis backend attached'

    def test__availability(self, loader: ConfigLoader):
        assert loader.get('name') == 'wroom-wroom'

    def test__default(self, loader: ConfigLoader):
        sentinel = object()
        assert loader.get('nonexi', default=sentinel) is sentinel

    def test__from_env__raises_on_empty_values(self):
        with pytest.raises(ValueError):
            ConfigLoader.from_env([], {})

    def test__from_env(self):
        env = {
            'CONFIG__PARSERS': 'EnvironmentParser,RedisParser,YamlParser',
            'ENVIRONMENTPARSER__SCOPE': 'nested',
            'YAMLPARSER__CONFIG': './tests/data/config.yml',
            'REDISPARSER__HOST': 'wtf.test',
            'NESTED__VARIABLE': 'i_am_here',
        }

        loader = ConfigLoader.from_env(env=env)
        assert [type(p) for p in loader.parsers] == [EnvironmentParser, RedisParser, YamlParser]
        assert loader.get('variable') == 'i_am_here', 'Ensure env copied from ConfigLoader'

        with pytest.raises(Exception):
            loader.get('nothing.here')

        loader = ConfigLoader.from_env(env=env, silent=True)
        assert loader.get('nothing.here', True) is True

        loader = ConfigLoader.from_env(parser_modules=['EnvironmentParser'], env={})
        assert loader.parsers

    def test__import_parsers(self):
        parsers = list(ConfigLoader.import_parsers([
            'EnvironmentParser',
            'django_docker_helpers.config.backends.YamlParser'
        ]))
        assert parsers == [EnvironmentParser, YamlParser]

    def test__load_parser_options_from_env(self):
        env = {
            'REDISPARSER__ENDPOINT': 'go.deep',
            'REDISPARSER__HOST': 'my-host',
            'REDISPARSER__PORT': '66',
        }

        res = ConfigLoader.load_parser_options_from_env(RedisParser, env)
        assert res == {'endpoint': 'go.deep', 'host': 'my-host', 'port': 66}

        env = {
            'ENVIRONMENTPARSER__SCOPE': 'deep',
        }
        res = ConfigLoader.load_parser_options_from_env(EnvironmentParser, env)
        assert res == {'scope': 'deep'}

    def test__config_read_queue(self,
                                loader: ConfigLoader,
                                store_mpt_consul_config,
                                store_mpt_redis_config,
                                store_consul_config,
                                store_redis_config):
        loader.get('some.variable')
        loader.get('some.brutal')
        loader.get('debug', coerce_type=bool)
        loader.get('i.am.redis')
        loader.get('variable')
        loader.get('name')
        loader.get('nothing.here', 'very long string lol')
        loader.get('secret')
        loader.get('something.long', list(range(100)))
        loader.get('something.long.q', '=' * 80)

        assert loader.config_read_queue
        assert '\033[0m' in loader.format_config_read_queue(use_color=True)
        assert '\033[0m' not in loader.format_config_read_queue(use_color=False)

        loader.print_config_read_queue(use_color=True)

    def test__shortcut_for_get_method(self, loader: ConfigLoader):
        assert loader('some.variable')

    def test__get_with_required(self, loader: ConfigLoader):
        assert loader.get('some.variable', required=True)

        with pytest.raises(exceptions.RequiredValueIsEmpty):
            loader.get('some.nonexistent_var', required=True)



import os

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
from yaml import dump as yaml_dump

from django_docker_helpers.config import exceptions
from django_docker_helpers.config.backends.consul_parser import ConsulParser

pytestmark = [pytest.mark.backend, pytest.mark.consul]

CONSUL_HOST = os.getenv('CONSUL_HOST', '127.0.0.1')
CONSUL_PORT = os.getenv('CONSUL_PORT', 8500)

SAMPLE = {
    'bool_flag': '',  # flag
    'unicode': 'вася',
    'none_value': None,
    'debug': True,
    'mixed': ['ascii', 'юникод', 1, {'d': 1}, {'b': 2}],
    'nested': {
        'a': {
            'b': 2
        }
    }
}


@pytest.fixture
def consul_parser():
    return ConsulParser('my/server/config.yml', host=CONSUL_HOST, port=CONSUL_PORT)


@pytest.fixture
def store_consul_config():
    import consul

    c = consul.Consul(host=CONSUL_HOST, port=CONSUL_PORT)
    data = yaml_dump(SAMPLE, allow_unicode=True).encode()
    c.kv.put('my/server/config.yml', data)

    c.kv.put('my/server/empty.yml', None)
    return c


# noinspection PyMethodMayBeStatic,PyUnusedLocal,PyShadowingNames
class ConsulBackendTest:
    def test__consul_parser__client(self, consul_parser):
        assert consul_parser.client

    def test__consul_parser__inner_parser(self, store_consul_config, consul_parser):
        assert consul_parser.client
        assert consul_parser.inner_parser

    def test__consul_parser__configs_equal(self, store_consul_config, consul_parser):
        assert consul_parser.inner_parser.data == SAMPLE

    def test__consul_parser__inner_parser__exceptions(self, store_consul_config):
        with pytest.raises(exceptions.KVStorageKeyDoestNotExist):
            c = ConsulParser('nothing/here', host=CONSUL_HOST, port=CONSUL_PORT)
            assert c.inner_parser

        with pytest.raises(exceptions.KVStorageValueIsEmpty):
            c = ConsulParser('my/server/empty.yml', host=CONSUL_HOST, port=CONSUL_PORT)
            assert c.inner_parser

    def test__consul_parser__get(self, store_consul_config, consul_parser: ConsulParser):
        assert consul_parser.get('nested.a.b', coerce_type=int) == 2
        assert consul_parser.get('nested.a') == {'b': 2}
        assert consul_parser.get('nested.nothing', default='qwe') == 'qwe'




import os

# noinspection PyPackageRequirements
import pytest

from django_docker_helpers.config.backends.mpt_consul_parser import MPTConsulParser
from django_docker_helpers.utils import mp_serialize_dict

pytestmark = [pytest.mark.backend, pytest.mark.consul]

CONSUL_HOST = os.getenv('CONSUL_HOST', '127.0.0.1')
CONSUL_PORT = os.getenv('CONSUL_PORT', 8500)


@pytest.fixture
def store_consul_config():
    import consul
    sample = {
        'bool_flag': '',  # flag
        'unicode': 'вася',
        'none_value': None,
        'debug': True,
        'mixed': ['ascii', 'юникод', 1, {'d': 1}, {'b': 2}],
        'nested': {
            'a': {
                'b': 2
            }
        }
    }
    c = consul.Consul(host=CONSUL_HOST, port=CONSUL_PORT)
    for path, value in mp_serialize_dict(sample, separator='/'):
        c.kv.put(path, value)


# noinspection PyMethodMayBeStatic,PyUnusedLocal,PyShadowingNames
class MPTConsulBackendTest:
    def test__mpt_consul_parser__get(self, store_consul_config):
        p = MPTConsulParser(host=CONSUL_HOST, port=CONSUL_PORT, path_separator='.')
        assert p.get('nested.a.b') == '2', 'Ensure value is a valid string if no coercers present'
        assert p.get('debug') == 'true', 'Ensure value is a valid string if no coercers present'

        # consul-python treats b'' as None, so we need to armor it in serializer
        assert p.get('bool_flag') == '', 'Ensure value is a valid string if no coercers present'
        assert p.get('bool_flag', coerce_type=bool) is True, 'Ensure value treated as True since it is a flag'

        assert p.get('nested.a.b', coerce_type=int) == 2, 'Ensure coercing works for simple types'
        assert p.get('debug', coerce_type=bool) is True, 'Ensure boolean works'

        assert p.get('unicode') == 'вася', 'Ensure unicode works fine'
        assert p.get('nested') is None, 'Ensure nested values are not returned'
        assert p.get('mixed') == ['ascii', 'юникод', 1, {'d': 1}, {'b': 2}]

    def test__mpt_consul_parser__scope(self, store_consul_config):
        p = MPTConsulParser(host=CONSUL_HOST, port=CONSUL_PORT, scope='nested', path_separator='.')
        assert p.get('a.b') == '2'



from time import sleep

from django.conf import settings
from django.core.cache import caches
from django.core.management import execute_from_command_line
from django.db import OperationalError, connections

from django_docker_helpers.utils import wf, run_env_once


@run_env_once
def ensure_caches_alive(max_retries: int = 100,
                        retry_timeout: int = 5,
                        exit_on_failure: bool = True) -> bool:
    """
    Checks every cache backend alias in ``settings.CACHES`` until it becomes available. After ``max_retries``
    attempts to reach any backend are failed it returns ``False``. If ``exit_on_failure`` is set it shuts down with
    ``exit(1)``.

    It sets the ``django-docker-helpers:available-check`` key for every cache backend to ensure
    it's receiving connections. If check is passed the key is deleted.

    :param exit_on_failure: set to ``True`` if there's no sense to continue
    :param int max_retries: a number of attempts to reach cache backend, default is ``100``
    :param int retry_timeout: a timeout in seconds between attempts, default is ``5``
    :return: ``True`` if all backends are available ``False`` if any backend check failed
    """
    for cache_alias in settings.CACHES.keys():
        cache = caches[cache_alias]
        wf('Checking redis connection alive for cache `%s`... ' % cache_alias, False)
        for i in range(max_retries):
            try:
                cache.set('django-docker-helpers:available-check', '1')
                assert cache.get('django-docker-helpers:available-check') == '1'
                cache.delete('django-docker-helpers:available-check')
                wf('[+]\n')
                break
            except Exception as e:
                wf(str(e) + '\n')
                sleep(retry_timeout)
        else:
            wf('Tried %s time(s). Shutting down.\n' % max_retries)
            exit_on_failure and exit(1)
            return False
    return True


@run_env_once
def ensure_databases_alive(max_retries: int = 100,
                           retry_timeout: int = 5,
                           exit_on_failure: bool = True) -> bool:
    """
    Checks every database alias in ``settings.DATABASES`` until it becomes available. After ``max_retries``
    attempts to reach any backend are failed it returns ``False``. If ``exit_on_failure`` is set it shuts down with
    ``exit(1)``.

    For every database alias it tries to ``SELECT 1``. If no errors raised it checks the next alias.

    :param exit_on_failure: set to ``True`` if there's no sense to continue
    :param int max_retries: number of attempts to reach every database; default is ``100``
    :param int retry_timeout: timeout in seconds between attempts
    :return: ``True`` if all backends are available, ``False`` if any backend check failed
    """
    template = """
    =============================
    Checking database connection `{CONNECTION}`:
        Engine: {ENGINE}
        Host: {HOST}
        Database: {NAME}
        User: {USER}
        Password: {PASSWORD}
    =============================\n"""
    for connection_name in connections:
        _db_settings = dict.fromkeys(['ENGINE', 'HOST', 'NAME', 'USER', 'PASSWORD'])
        _db_settings.update(settings.DATABASES[connection_name])
        _db_settings['CONNECTION'] = connection_name
        if _db_settings.get('PASSWORD'):
            _db_settings['PASSWORD'] = 'set'

        wf(template.format(**_db_settings))
        wf('Checking db connection alive... ', False)

        for i in range(max_retries):
            try:
                cursor = connections[connection_name].cursor()
                cursor.execute('SELECT 1')
                cursor.fetchone()

                wf('[+]\n')
                break
            except OperationalError as e:
                wf(str(e))
                sleep(retry_timeout)
        else:
            wf('Tried %s time(s). Shutting down.\n' % max_retries)
            exit_on_failure and exit(1)
            return False
    return True


@run_env_once
def migrate() -> bool:
    """
    Runs Django migrate command.

    :return: always ``True``
    """
    wf('Applying migrations... ', False)
    execute_from_command_line(['./manage.py', 'migrate'])
    wf('[+]\n')
    return True


@run_env_once
def modeltranslation_sync_translation_fields() -> bool:
    """
    Runs ``modeltranslation``'s ``sync_translation_fields`` manage.py command:
    ``execute_from_command_line(['./manage.py', 'sync_translation_fields', '--noinput'])``

    :return: ``None`` if modeltranslation is not specified is ``INSTALLED_APPS``, ``True`` if all synced.
    """
    # if modeltranslation present ensure it's migrations applied too

    if 'modeltranslation' in settings.INSTALLED_APPS:
        wf('Applying translations for models... ', False)
        execute_from_command_line(['./manage.py', 'sync_translation_fields', '--noinput'])
        wf('[+]\n')
        return True
